<?php

namespace SGalinski\SgJobs\Property\TypeConverter;

/***************************************************************
 *  Copyright notice
 *
 *  (c) 2014 Helmut Hummel
 *
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 ***************************************************************/

use TYPO3\CMS\Core\Resource\DuplicationBehavior;
use TYPO3\CMS\Core\Resource\File as FalFile;
use TYPO3\CMS\Core\Resource\FileReference as FalFileReference;
use TYPO3\CMS\Core\Resource\ResourceFactory;
use TYPO3\CMS\Core\Utility\GeneralUtility;
use TYPO3\CMS\Core\Utility\PathUtility;
use SGalinski\SgJobs\Domain\Model\FileReference;
use TYPO3\CMS\Extbase\Error\Error;
use TYPO3\CMS\Extbase\Object\ObjectManagerInterface;
use TYPO3\CMS\Extbase\Persistence\Generic\PersistenceManager;
use TYPO3\CMS\Extbase\Property\Exception\TypeConverterException;
use TYPO3\CMS\Extbase\Property\PropertyMappingConfigurationInterface;
use TYPO3\CMS\Extbase\Property\TypeConverterInterface;
use TYPO3\CMS\Extbase\Security\Cryptography\HashService;

/**
 * Class UploadedFileReferenceConverter
 */
class UploadedFileReferenceConverter implements TypeConverterInterface {

	/**
	 * Folder where the file upload should go to (including storage).
	 *
	 * @var string
	 */
	protected $uploadFolder = '1:/user_upload/';

	/**
	 * How to handle a upload when the name of the uploaded file conflicts.
	 *
	 * @var string
	 */
	protected $uploadConflictMode = DuplicationBehavior::REPLACE;

	/**
	 * Whether to replace an already present resource.
	 * Useful for "maxitems = 1" fields and properties
	 * with no ObjectStorage annotation.
	 *
	 * @var string
	 */
	protected $allowedFileExtensions = '';

	/**
	 * @var array<string>
	 */
	protected $sourceTypes = ['array'];

	/**
	 * @var string
	 */
	protected $targetType = FileReference::class;

	/**
	 * Take precedence over the available FileReferenceConverter
	 *
	 * @var int
	 */
	protected $priority = 30;

	/**
	 * @var ResourceFactory
	 */
	protected $resourceFactory;

	/**
	 * Inject the ResourceFactory
	 *
	 * @param ResourceFactory $resourceFactory
	 */
	public function injectResourceFactory(ResourceFactory $resourceFactory) {
		$this->resourceFactory = $resourceFactory;
	}

	/**
	 * @var HashService
	 */
	protected $hashService;

	/**
	 * Inject the HashService
	 *
	 * @param HashService $hashService
	 */
	public function injectHashService(HashService $hashService) {
		$this->hashService = $hashService;
	}

	/**
	 * @var PersistenceManager
	 */
	protected $persistenceManager;

	/**
	 * Inject the PersistenceManager
	 *
	 * @param PersistenceManager $persistenceManager
	 */
	public function injectPersistenceManager(PersistenceManager $persistenceManager) {
		$this->persistenceManager = $persistenceManager;
	}

	/**
	 * @var \TYPO3\CMS\Core\Resource\FileInterface[]
	 */
	protected $convertedResources = [];

	/**
	 * @var string
	 */
	protected $targetUploadFileName = '';

	/**
	 * @var ObjectManagerInterface
	 */
	protected $objectManager;

	/**
	 * @param ObjectManagerInterface $objectManager
	 */
	public function injectObjectManager(ObjectManagerInterface $objectManager) {
		$this->objectManager = $objectManager;
	}

	/**
	 * Returns the list of source types the TypeConverter can handle.
	 * Must be PHP simple types, classes or object is not allowed.
	 *
	 * @return array<string>
	 * @api
	 */
	public function getSupportedSourceTypes(): array {
		return $this->sourceTypes;
	}

	/**
	 * Return the target type this TypeConverter converts to.
	 * Can be a simple type or a class name.
	 *
	 * @return string
	 * @api
	 */
	public function getSupportedTargetType(): string {
		return $this->targetType;
	}

	/**
	 * Returns the $originalTargetType unchanged in this implementation.
	 *
	 * @param mixed $source the source data
	 * @param string $originalTargetType the type we originally want to convert to
	 * @param PropertyMappingConfigurationInterface $configuration
	 * @return string
	 * @api
	 */
	public function getTargetTypeForSource(
		$source, $originalTargetType,
		PropertyMappingConfigurationInterface $configuration = NULL
	): string {
		return $originalTargetType;
	}

	/**
	 * Return the priority of this TypeConverter. TypeConverters with a high priority are chosen before low priority.
	 *
	 * @return int
	 * @api
	 */
	public function getPriority(): int {
		return $this->priority;
	}

	/**
	 * This implementation always returns TRUE for this method.
	 *
	 * @param mixed $source the source data
	 * @param string $targetType the type to convert to.
	 * @return bool TRUE if this TypeConverter can convert from $source to $targetType, FALSE otherwise.
	 * @api
	 */
	public function canConvertFrom($source, $targetType): bool {
		return TRUE;
	}

	/**
	 * Returns an empty list of sub property names
	 *
	 * @param mixed $source
	 * @return array
	 * @api
	 */
	public function getSourceChildPropertiesToBeConverted($source): array {
		return [];
	}

	/**
	 * This method is never called, as getSourceChildPropertiesToBeConverted() returns an empty array.
	 *
	 * @param string $targetType
	 * @param string $propertyName
	 * @param PropertyMappingConfigurationInterface $configuration
	 * @api
	 */
	public function getTypeOfChildProperty(
		$targetType, $propertyName, PropertyMappingConfigurationInterface $configuration
	) {
	}

	/**
	 * @return string
	 */
	public function getTargetUploadFileName(): string {
		return $this->targetUploadFileName;
	}

	/**
	 * @param string $targetUploadFileName
	 * @return void
	 */
	public function setTargetUploadFileName($targetUploadFileName) {
		$this->targetUploadFileName = $targetUploadFileName;
	}

	/**
	 * @return string
	 */
	public function getUploadFolder(): string {
		return $this->uploadFolder;
	}

	/**
	 * @param string $uploadFolder
	 * @return void
	 */
	public function setUploadFolder($uploadFolder) {
		$this->uploadFolder = $uploadFolder;
	}

	/**
	 * @return string
	 */
	public function getUploadConflictMode(): string {
		return $this->uploadConflictMode;
	}

	/**
	 * @param string $uploadConflictMode
	 * @return void
	 */
	public function setUploadConflictMode($uploadConflictMode) {
		$this->uploadConflictMode = $uploadConflictMode;
	}

	/**
	 * @return string
	 */
	public function getAllowedFileExtensions(): string {
		return $this->allowedFileExtensions;
	}

	/**
	 * @param string $allowedFileExtensions
	 * @return void
	 */
	public function setAllowedFileExtensions($allowedFileExtensions) {
		$this->allowedFileExtensions = $allowedFileExtensions;
	}

	/**
	 * Actually convert from $source to $targetType, taking into account the fully
	 * built $convertedChildProperties and $configuration.
	 *
	 * @param array $source
	 * @param string $targetType
	 * @param array $convertedChildProperties
	 * @param PropertyMappingConfigurationInterface $configuration
	 * @return null|FileReference|Error
	 * @api
	 */
	public function convertFrom(
		$source, $targetType, array $convertedChildProperties = [],
		PropertyMappingConfigurationInterface $configuration = NULL
	) {
		if ($source['name'] === '' && \is_array($source['submittedFile'])) {
			$source = $source['submittedFile'];
			$source['wasUploaded'] = TRUE;
		}

		if ($source['error'] === \UPLOAD_ERR_NO_FILE) {
			return NULL;
		}

		if ($source['error'] !== '' && (int) $source['error'] !== \UPLOAD_ERR_OK) {
			switch ($source['error']) {
				case \UPLOAD_ERR_INI_SIZE:
				case \UPLOAD_ERR_FORM_SIZE:
				case \UPLOAD_ERR_PARTIAL:
					return new Error('Error Code: ' . $source['error'], 1264440823);
				default:
					return new Error(
						'An error occurred while uploading. Please try again or contact the administrator if the problem remains',
						1340193849
					);
			}
		}

		if (isset($this->convertedResources[$source['tmp_name']])) {
			return $this->convertedResources[$source['tmp_name']];
		}

		$resource = NULL;

		try {
			if ($configuration) {
				$resource = $this->importUploadedResource($source);
				$this->convertedResources[$source['tmp_name']] = $resource;
			}
		} catch (\Exception $e) {
			return new Error($e->getMessage(), $e->getCode());
		}

		return $resource;
	}

	/**
	 * Import a resource and respect configuration given for properties
	 *
	 * @param array $uploadInfo
	 * @return FileReference|NULL
	 * @throws \InvalidArgumentException
	 * @throws TypeConverterException
	 * @throws \Exception
	 * @throws \TYPO3\CMS\Core\Resource\Exception\ExistingTargetFolderException
	 * @throws \TYPO3\CMS\Core\Resource\Exception\InsufficientFolderAccessPermissionsException
	 * @throws \TYPO3\CMS\Core\Resource\Exception\InsufficientFolderWritePermissionsException
	 */
	protected function importUploadedResource(array $uploadInfo) {
		$uploadedFile = NULL;
		$filePathInfo = PathUtility::pathinfo($uploadInfo['name']);
		$fileName = $this->getTargetUploadFileName();
		$finalFileName = $fileName . '.' . strtolower($filePathInfo['extension']);
		if (!$uploadInfo['wasUploaded']) {
			if (!GeneralUtility::verifyFilenameAgainstDenyPattern($uploadInfo['name'])) {
				throw new TypeConverterException(
					'Uploading files with PHP file extensions is not allowed!', 1399312430
				);
			}

			$allowedFileExtensions = $this->getAllowedFileExtensions();
			if (!GeneralUtility::inList($allowedFileExtensions, strtolower($filePathInfo['extension']))) {
				throw new TypeConverterException('File extension is not allowed!', 1399312430);
			}

			$uploadFolderId = $this->getUploadFolder();
			$resourceFactory = $this->objectManager->get(ResourceFactory::class);
			$storage = $resourceFactory->getStorageObject(1);
			$folderWithoutStorage = preg_replace('/^1:/', '', $uploadFolderId);
			if (!$storage->hasFolder($folderWithoutStorage)) {
				$storage->createFolder($folderWithoutStorage);
			}

			$uploadFolder = $this->resourceFactory->retrieveFileOrFolderObject($uploadFolderId);
			$uploadedFile = $uploadFolder->addUploadedFile($uploadInfo, $this->getUploadConflictMode());
			if ($fileName !== '') {
				$uploadedFile->rename($finalFileName, $this->getUploadConflictMode());
			}

		} else {
			$uploadFolderId = $this->getUploadFolder();
			// Security protection to not allow manipulations outside of this specific folder
			if (strpos($uploadFolderId, '/JobApplication/') !== FALSE) {
				try {
					$uploadedFile = $this->resourceFactory->retrieveFileOrFolderObject(
						$uploadFolderId . '/' . $finalFileName
					);
				} catch (\Exception $exception) {
					// nope
				}

			} else {
				throw new \RuntimeException('Not allowed!');
			}
		}

		if (!$uploadedFile) {
			return NULL;
		}

		return $this->createFileReferenceFromFalFileObject($uploadedFile);
	}

	/**
	 * @param FalFile $file
	 * @return FileReference
	 */
	protected function createFileReferenceFromFalFileObject(FalFile $file
	): FileReference {
		$fileReference = $this->resourceFactory->createFileReferenceObject(
			[
				'uid_local' => $file->getUid(),
				'uid_foreign' => uniqid('NEW_', TRUE),
				'uid' => uniqid('NEW_', TRUE),
				'crop' => NULL,
			]
		);
		return $this->createFileReferenceFromFalFileReferenceObject($fileReference);
	}

	/**
	 * @param FalFileReference $falFileReference
	 * @return FileReference
	 */
	protected function createFileReferenceFromFalFileReferenceObject(FalFileReference $falFileReference
	): FileReference {
		$fileReference = $this->objectManager->get(FileReference::class);
		$fileReference->setOriginalResource($falFileReference);
		return $fileReference;
	}
}
